@@ -1,6 +1,8 @@
 #
 # Debatable flask-side auth management
 #
+# This code is public domain.
+#
 
 from typing import Optional, Union, Callable, Dict, List, Any
 from flask import Flask, request, Response
@@ -27,9 +29,23 @@ class AuthException(BaseException):
 APP: Optional[Flask] = None
 CONF: Optional[Dict[str, Any]] = None
 
+# auth type
+AUTH: Optional[str] = None
+LAZY: Optional[bool] = None
+
 # auth token
+NAME: Optional[str] = None
 REALM: Optional[str] = None
 SECRET: Optional[str] = None
+DELAY: Optional[int] = None
+GRACE: Optional[int] = None
+HASH: Optional[str] = None
+SIGLEN: Optional[int] = None
+
+# parameter names
+LOGIN: Optional[str] = None
+USERP: Optional[str] = None
+PASSP: Optional[str] = None
 
 # password management
 PM: Optional[CryptContext] = None
@@ -57,23 +73,44 @@ def auth_cleanup(res: Response):
 def setConfig(app: Flask,
               gup: GetUserPasswordType = None,
               uig: UserInGroupType = None):
-    global APP, CONF, REALM, SECRET, PM, get_user_password, user_in_group
+    global APP, CONF, AUTH, LAZY
+    global NAME, REALM, SECRET, DELAY, GRACE, HASH, SIGLEN
+    global LOGIN, USERP, PASSP, PM
+    global get_user_password, user_in_group
     # overall setup
     APP = app
     CONF = app.config
     app.after_request(auth_cleanup)
+    AUTH = CONF.get("FSA_TYPE", "httpd")
+    LAZY = CONF.get("FSA_LAZY", True)
     # token setup
+    NAME = CONF.get("FSA_TOKEN_NAME", "auth")
     import re
     realm = CONF.get("FSA_TOKEN_REALM", app.name).lower()
     # tr -cd "[a-z0-9_]" "": is there a better way to do that?
-    REALM = "".join(c for c in realm if re.match("[a-z0-9_]", c))
+    REALM = "".join(c for c in realm if re.match("[-a-z0-9_]", c))
     import random
     import string
     # list of 94 chars, about 6.5 bits per char
     chars = string.ascii_letters + string.digits + string.punctuation
-    SECRET = CONF["FSA_TOKEN_SECRET"] if "FSA_TOKEN_SECRET" in CONF else \
-        ''.join(random.SystemRandom().choices(chars, k=64))  # > 256 bits
+    if "FSA_TOKEN_SECRET" in CONF:
+        SECRET = CONF["FSA_TOKEN_SECRET"]
+        if SECRET is not None and len(SECRET) < 16:
+            log.warning("token secret is short")
+    else:
+        log.warning("random token secret, only ok for one process app")
+        SECRET = ''.join(random.SystemRandom().choices(chars, k=40))
+    DELAY = CONF.get("FSA_TOKEN_DELAY", 60)
+    GRACE = CONF.get("FSA_TOKEN_GRACE", 0)
+    HASH = CONF.get("FSA_TOKEN_HASH", "blake2s")
+    SIGLEN = CONF.get("FSA_TOKEN_LENGTH", 16)
+    # parameters
+    LOGIN = CONF.get("FSA_FAKE_LOGIN", "LOGIN")
+    USERP = CONF.get("FSA_PARAM_USER", "USER")
+    PASSP = CONF.get("FSA_PARAM_PASS", "PASS")
     # password setup
+    # passlib context is a pain, you have to know the scheme name to set its
+    # round which make it impossible to configure directly.
     scheme = CONF.get("FSA_PASSWORD_SCHEME", "bcrypt")
     options = CONF.get("FSA_PASSWORD_OPTIONS", {'bcrypt__default_rounds': 4})
     PM = CryptContext(schemes=[scheme], **options)
@@ -94,7 +131,7 @@ def get_fake_auth():
     assert request.environ["REMOTE_ADDR"][:4] == "127.", \
         "fake auth only on localhost"
     params = request.values if request.json is None else request.json
-    user = params.get(CONF.get("FSA_FAKE_LOGIN", "LOGIN"), None)
+    user = params.get(LOGIN, None)
     # it could check that the user exists in db
     if user is None:
         raise AuthException("missing login parameter", 401)
@@ -111,9 +148,6 @@ def get_fake_auth():
 # note: passlib bcrypt is Apache compatible
 #
 
-# passlib context is a pain, you have to know the scheme name to set its round
-# which make it impossible to configure directly.
-
 # verify password
 def check_password(user, pwd, ref):
     if not PM.verify(pwd, ref):
@@ -168,13 +202,11 @@ def get_basic_auth():
 def get_param_auth():
     assert request.remote_user is None
     params = request.values if request.json is None else request.json
-    userp = CONF.get("FSA_PARAM_USER", "USER")
-    pwdp = CONF.get("FSA_PARAM_PASS", "PASS")
-    user, pwd = params.get(userp, None), params.get(pwdp, None)
+    user, pwd = params.get(USERP, None), params.get(PASSP, None)
     if user is None:
-        raise AuthException(f"missing login parameter: {userp}", 401)
+        raise AuthException(f"missing login parameter: {USERP}", 401)
     if pwd is None:
-        raise AuthException(f"missing password parameter: {pwdp}", 401)
+        raise AuthException(f"missing password parameter: {PASSP}", 401)
     if not request.is_secure:
         log.warning("password authentication over an insecure request")
     check_db_password(user, pwd)
@@ -194,6 +226,7 @@ def get_param_auth():
 # FSA_TOKEN_HASH: hashlib algorithm for token authentication ("blake2s")
 # FSA_TOKEN_LENGTH: number of signature bytes (32)
 # FSA_TOKEN_DELAY: token validity in minutes (60)
+# FSA_TOKEN_GRACE: grace delay for token validity in minutes (0)
 # FSA_TOKEN_SECRET: signature secret for tokens (mandatory!)
 # FSA_TOKEN_REALM: token realm (lc simplified app name)
 #
@@ -201,10 +234,9 @@ def get_param_auth():
 # sign data with secret
 def compute_signature(data, secret):
     import hashlib
-    h = hashlib.new(CONF.get("FSA_TOKEN_HASH", "blake2s"))
-    siglen = CONF.get("FSA_TOKEN_LENGTH", 32)
+    h = hashlib.new(HASH)
     h.update(f"{data}:{secret}".encode())
-    return h.digest()[:siglen].hex()
+    return h.digest()[:SIGLEN].hex()
 
 
 # build a timestamp string
@@ -222,8 +254,7 @@ def compute_token(realm, user, delay, secret):
 
 # create a new token for user depending on the configuration
 def create_token(user):
-    delay = CONF.get("FSA_TOKEN_DELAY", 60)
-    return compute_token(REALM, user, delay, SECRET)
+    return compute_token(REALM, user, DELAY, SECRET)
 
 
 # tell whether token is ok: return validated user or None
@@ -235,13 +266,12 @@ def get_token_auth(token):
         log.info(f"LOGIN (token): unexpected realm {realm}")
         raise AuthException(f"unexpected realm: {realm}", 401)
     # check signature
-    ref = compute_signature(f"{realm}:{user}:{limit}",
-                            CONF["FSA_TOKEN_SECRET"])
+    ref = compute_signature(f"{realm}:{user}:{limit}", SECRET)
     if ref != sig:
         log.info("LOGIN (token): invalid signature")
         raise AuthException("invalid auth token signature", 401)
-    # check limit
-    now = get_timestamp(dt.datetime.utcnow())
+    # check limit with a grace time
+    now = get_timestamp(dt.datetime.utcnow() - dt.timedelta(minutes=GRACE))
     if now > limit:
         log.info("LOGIN (token): token {token} has expired")
         raise AuthException("expired auth token", 401)
@@ -256,35 +286,32 @@ def get_user():
     global USER
     USER = None
 
-    # default is to rely on the http server provided authentication
-    auth_type = CONF.get("FSA_TYPE", "httpd")
+    if AUTH is None:
+        raise AuthException("FlaskSimpleAuth module not initialized", 500)
 
-    if auth_type == "httpd":
+    if AUTH == "httpd":
 
         log.info(f"LOGIN (httpd): {request.remote_user}")
         USER = request.remote_user
 
-    elif auth_type in ("fake", "param", "basic", "token", "password"):
+    elif AUTH in ("fake", "param", "basic", "token", "password"):
 
-        # whether to attempt token auth
-        token_name = CONF.get("FSA_TOKEN_NAME", "auth")
-
-        # first look for the token
+        # check for token
         if SECRET is not None and SECRET != "":
             params = request.values if request.json is None else request.json
-            token = params.get(token_name, None)
+            token = params.get(NAME, None)
             if token is not None:
                 USER = get_token_auth(token)
 
-        # else try other auth schemes
+        # else try other schemes
         if USER is None:
-            if auth_type == "param":
+            if AUTH == "param":
                 USER = get_param_auth()
-            elif auth_type == "basic":
+            elif AUTH == "basic":
                 USER = get_basic_auth()
-            elif auth_type == "fake":
+            elif AUTH == "fake":
                 USER = get_fake_auth()
-            elif auth_type == "password":
+            elif AUTH == "password":
                 try:
                     USER = get_basic_auth()
                 except AuthException:  # try param
@@ -292,16 +319,16 @@ def get_user():
             else:
                 raise AuthException("auth token is required", 401)
 
-        assert USER is not None  # otherwise a exception would have been raised
+        assert USER is not None  # else an exception would have been raised
         return USER
 
     else:
 
-        raise AuthException(f"unexpected FSA_TYPE: {auth_type}", 500)
+        raise AuthException(f"unexpected authentication type: {AUTH}", 500)
 
 
 #
-# authorization helper
+# authorize decorator
 #
 class authorize:
 
@@ -316,17 +343,21 @@ class authorize:
 
     def __call__(self, fun):
         def wrapper(*args, **kwargs):
+            global USER
             if USER is None:
-                return "", 401
-            permitted = False
+                # no current user, try to get one?
+                if LAZY:
+                    try:
+                        USER = get_user()
+                    except AuthException:
+                        return "", 401
+                else:
+                    return "", 401
             for g in self.groups:
                 if user_in_group(USER, g):
-                    permitted = True
-                    break
-            if not permitted:
-                return "", 403
-            else:
-                return fun(*args, **kwargs)
-        # work around flask reliance on the function name
+                    return fun(*args, **kwargs)
+            # else no matching group
+            return "", 403
+        # work around flask unwitty reliance on the function name
         wrapper.__name__ = fun.__name__
         return wrapper
